// linesocket.cpp
// Copyright (c) 2008 Phillip Hellewell

#include "lineclient.h"
#include "utils.h"
#include <ogc/lwp.h>
#include <network.h>

static lwp_t g_listen_thread = 0;

using namespace std;

LineClient::LineClient()
	: m_socket(INVALID_SOCKET)
	, m_shutdown(false)
	, m_connected(false)
	, m_index1(0)
	, m_index2(0)
{
}

LineClient::~LineClient()
{
	m_shutdown = true;
	disconnect();
}

bool LineClient::init()
{
	// Return if listen thread already initialized.
	if( g_listen_thread )
		return true;

	// Start the listen thread.
	return LWP_CreateThread(&g_listen_thread, static_listenloop, this, NULL, 0, 80) >= 0;
}

bool LineClient::connect(const string& host, int port)
{
	// Close any previous connection.
	disconnect();

	// Create socket.
	m_socket = net_socket(AF_INET, SOCK_STREAM, 0);
	if(m_socket == INVALID_SOCKET)
		return false;

	// Get server IP address.  First attempt host as an ip address.
	u32 ipaddr = inet_addr(host.c_str());

	// If that didn't work, try host as a name.
	if( ipaddr == 0xFFFFFFFF ) {
		struct hostent* hent = net_gethostbyname((char*)host.c_str());
		if( !hent )
			return false;
		memcpy(&ipaddr, hent->h_addr_list[0], 4);
	}

    // Fill address struct and connect to server.
	sockaddr_in addr = { 0 };
	addr.sin_family			= AF_INET;
	addr.sin_port			= htons(port);
	addr.sin_addr.s_addr	= ipaddr;
	if( net_connect(m_socket, (sockaddr*)&addr, sizeof(sockaddr_in)) != 0 )
		return false;

	m_connected = true;

	return m_connected;
}

bool LineClient::disconnect()
{
	m_connected = false;

	if( m_socket == INVALID_SOCKET )
		return true;

	if( net_shutdown(m_socket, 0) != 0 || net_close(m_socket) != 0 )
		return false;

	m_socket = INVALID_SOCKET;
	return true;
}

bool LineClient::sendLine(const string& line)
{
	if( !connected() )
		return false;

	string line2        = line + '\n';
	const char* linestr = line2.c_str();
	int length          = line2.length();
	int bytessent       = 0;

	// Send line.
	while( bytessent < length)
	{
		int bsent = net_send(m_socket, &linestr[bytessent], length - bytessent, 0);
		if( bsent == SOCKET_ERROR )
		{
#ifdef VERBOSE
			printf("ERROR SENDING: %s\n", line.c_str());
			printf("DISCONNECTING.\n");
#endif
			disconnect();
			return false;
		}
		bytessent += bsent;
	}

#ifdef VERBOSE
	printf("\nSENT: %s\n", line.c_str());
#endif

	return true;
}

bool LineClient::getLine(string* line)
{
	// Wait for the line to be received if it hasn't been already.
	while( connected() && !hasLine() )
		millisleep(20);

	if( !hasLine() )
		return false;

	// Line is ready to go!
	if( line )
		*line = m_line[m_index1];

#ifdef VERBOSE
	printf("RECEIVED: %s\n", m_line[m_index1].c_str());
#endif

	m_index1++;
	m_index1 %= LINE_BUFSIZE;

	return true;
}

void* LineClient::static_listenloop(void* line_client)
{
	reinterpret_cast<LineClient*>(line_client)->listenloop();
	return NULL;
}

void LineClient::listenloop()
{
	string line;
	while( !m_shutdown )
	{
		if( !connected() )
		{
			line.clear();
			millisleep(10);
		}
		else
		{
			// Receive line(s) here.
			char buf[1000];
			s32 got = net_recv(m_socket, buf, sizeof(buf), 0);
			if( got <= 0 )
			{
#ifdef VERBOSE
				printf("GOT DISCONNECTED!!!\n");
#endif
				disconnect();
				return;
			}

			// Add line(s) to line buffer.
			for( s32 i = 0; i < got; ++i )
			{
				if( buf[i] == '\n' )
				{
					addLine(line);
					line.clear();
				}
				else if( buf[i] != '\r' && buf[i] != '\0' )
				{
					line += buf[i];
				}
			}
		}
	}
}

void LineClient::addLine(const string& line)
{
	// Wait here forever while buffer is full.
	int newindex = (m_index2+1) % LINE_BUFSIZE;
	while( newindex == m_index1 )
		millisleep(10);

	// Add the line.
	m_line[m_index2] = line;

	// Let the caller know he can get a line now.
	m_index2 = newindex;
}

// eof
