// WiiMPC
// Copyright (c) 2008 Phillip Hellewell
#include <errno.h>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <stdio.h>
#include <gccore.h>
#include <network.h>
#include <fat.h>
#include <wiiuse/wpad.h>
#include "GRRLIB/GRRLIB.h"
#include "GRRLIB/fonts/GRRLIB_font1.h"
#include "mpclient.h"
#include "img/logo.h"
#include "img/background.h"
#include "img/timebar.h"
#include "img/play.h"
#include "img/pause.h"
#include "img/stop.h"
#include "img/volumebar.h"
#include "img/volumepos.h"

// Constants.
static const std::string CONF_FILENAME = "wiimpc.conf";
static const std::string DEFAULT_HOST = "mpdserver";
static const int PORT = 6600;
static const int GRRLIB_IS_DUMB_LOOP_COUNT = 3;
static const size_t MAX_WIDTH_CHARS = 38;

// Globals.
static std::string g_host;
MPClient g_client;
Mtx GXmodelView2D;
u8* g_font;
u8* g_background;
u8* g_logo;
u8* g_timebar;
u8* g_timepos_play;
u8* g_timepos_pause;
u8* g_timepos_stop;
u8* g_volumebar;
u8* g_volumepos;

struct StatusInfo
{
	StatusInfo() : isvalid(false), volume(0), songid(-1), time_elapsed(0), time_total(0) {}
	bool isvalid;
	int volume;
	std::string state;
	int songid;
	int time_elapsed;
	int time_total;
} g_status;

struct SongInfo
{
	SongInfo() : isvalid(false), songid(-1) {}
	bool isvalid;
	int songid;
	std::string filename;
	std::string artist;
	std::string title;
	int scroll;
} g_songinfo;

static void paint_background()
{
	GRRLIB_DrawImg(0, 0, 640, 480, g_background, 0.0f, 1.0f, 1.0f, 255.0f);
	GRRLIB_DrawImg(520, 420, 112, 36, g_logo, 0.0f, 1.0f, 1.0f, 255.0f);
}

static void paint_first()
{
	for( int i = 0; i < GRRLIB_IS_DUMB_LOOP_COUNT; ++i ) {
		paint_background();
		GRRLIB_Render();
	}
}

static void paint_netinit()
{
	for( int i = 0; i < GRRLIB_IS_DUMB_LOOP_COUNT; ++i ) {
		paint_background();
		GRRLIB_Printf(20, 40, g_font, 0xFFFFFFFF, 2, "Initializing network, please wait...");
		GRRLIB_Render();
	}
}

static void paint_status()
{
	if( !g_status.isvalid )
		return;

	u8* timepos;
	if( g_status.state == "play" )
		timepos = g_timepos_play;
	else if( g_status.state == "pause" )
		timepos = g_timepos_pause;
	else //if( g_status.state == "stop" )
		timepos = g_timepos_stop;

	int pos = g_status.time_elapsed * 600 / g_status.time_total;
	GRRLIB_DrawImg(20, 100, 600, 4, g_timebar, 0.0f, 1.0f, 1.0f, 255.0f);
	GRRLIB_DrawImg(20 + pos, 91, 20, 20, timepos, 0.0f, 1.0f, 1.0f, 255.0f);

	int min_tot = g_status.time_total / 60;
	int sec_tot = g_status.time_total % 60;
	int xpos = min_tot > 9 ? 540 : 560;
	GRRLIB_Printf(xpos, 120, g_font, 0xFFA0A0A0, 2, "%d:%02d", min_tot, sec_tot);

	int volpos = g_status.volume;
	if( volpos < 0 ) volpos = 0;
	if( volpos > 100 ) volpos = 100;
	GRRLIB_DrawImg(30, 160, 4, 100, g_volumebar, 0.0f, 1.0f, 1.0f, 255.0f);
	GRRLIB_DrawImg(21, 260 - 4 - volpos, 20, 8, g_volumepos, 0.0f, 1.0f, 1.0f, 255.0f);
}

// Return two x deltas for where the song name should be painted.
static void get_scroll_positions(size_t chars, int scroll, int* pos1, int* pos2)
{
	int width = static_cast<int>(chars) * 8 * 2;
	int period = width + 320;
	scroll %= period;
	*pos1 = 20 - scroll;
	*pos2 = *pos1 + period;
}

static std::string replace_all(const std::string& in, const std::string& sold, const std::string& snew)
{
     std::string ret = in;
     std::string::size_type pos = 0, lpos = 0;
     while( (pos = ret.find(sold, lpos)) != std::string::npos ) {
           ret.replace(pos, sold.size(), snew);
           lpos = pos + snew.size();
     }
     return ret;
}

// Prettify filename by removing extension, slashes, and underscores.
static std::string make_pretty_filename(const std::string& filename)
{
	std::string ret = filename;

	// Remove extension.
	if( ret.size() >= 4 && ret[ret.size()-4] == '.' )
		ret.resize(ret.size()-4);

	// Remove underscores.
	std::replace(ret.begin(), ret.end(), '_', ' ');

	// Remove slashes.
	ret = replace_all(ret, "/", " : ");

	return ret;
}

static void paint_songinfo()
{
	if( !g_songinfo.isvalid )
		return;

	// Piece name together from artist, title, and/or filename.
	std::string name;
	if( !g_songinfo.title.empty() ) {
		if( !g_songinfo.artist.empty() )
			name = g_songinfo.artist + " - ";
		name += g_songinfo.title;
	}
	if( name.empty() )
		name = make_pretty_filename(g_songinfo.filename);
	if( name.empty() )
		name = "(unknown)";

	if( name.size() <= MAX_WIDTH_CHARS ) {
		GRRLIB_Printf(20, 60, g_font, 0xFFFFFF40, 2, "%s", name.c_str());
	} else {
		int pos1, pos2;
		get_scroll_positions(name.size(), g_songinfo.scroll, &pos1, &pos2);
		GRRLIB_Printf(pos1, 60, g_font, 0xFFFFFF40, 2, "%s", name.c_str());
		GRRLIB_Printf(pos2, 60, g_font, 0xFFFFFF40, 2, "%s", name.c_str());
	}
}

static void paint()
{
	paint_background();

	// Draw main stuff.
	if( g_client.connected() ) {
		paint_status();
		paint_songinfo();
	} else {
		std::string msg = "Connecting to " + g_host + "...";
		GRRLIB_Printf(20, 40, g_font, 0xFFFFFFFF, 2, msg.c_str());
	}

	// Let the user know the WiiMote is disconnected by fading out the screen.
	s32 ret = WPAD_Probe(0, NULL);
	if( ret == WPAD_ERR_NO_CONTROLLER )
		GRRLIB_FillScreen(0x80000000);
}

static void paint_exit()
{
	for( int i = 0; i < GRRLIB_IS_DUMB_LOOP_COUNT; ++i ) {
		paint();
		GRRLIB_FillScreen(0x80000000);
		GRRLIB_Printf(20, 430, g_font, 0xFFFFFFFF, 2, "Exiting, please wait...");
		GRRLIB_Render();
	}
}

static void fail_warn(const char* message, int secs = 2)
{
	for( int i = 0; i < secs*60; ++i ) {
		paint_background();
		GRRLIB_Printf(20, 40, g_font, 0xFFFF4040, 2, message);
		GRRLIB_Render();
	}
}

static void fail_exit(const char* message, int secs = 2)
{
	fail_warn(message, secs);
	exit(0);
}

static s32 initialize_network()
{
    s32 ret;
    while( (ret = net_init()) == -EAGAIN )
		continue;
    return ret;
}

static bool is_playing()
{
	return g_status.isvalid && g_status.state == "play";
}

// Returns false if program should exit.
static bool process_wpad_commands()
{
	u32 wpaddown = WPAD_ButtonsDown(0) | WPAD_ButtonsDown(1) |
	               WPAD_ButtonsDown(2) | WPAD_ButtonsDown(3);

	if( wpaddown & WPAD_BUTTON_HOME )
		return false;

	// Don't send commands when not connected.
	if( !g_client.connected() )
		return true;

	std::string cmd;
	if( wpaddown & WPAD_BUTTON_A )
		cmd = is_playing() ? "pause" : "play";
	else if( wpaddown & WPAD_BUTTON_2 )
		cmd = "stop";
	else if( wpaddown & WPAD_BUTTON_LEFT )
		cmd = "previous";
	else if( wpaddown & WPAD_BUTTON_RIGHT )
		cmd = "next";
	else if( wpaddown & WPAD_BUTTON_MINUS )
		cmd = "volume -5";
	else if( wpaddown & WPAD_BUTTON_PLUS )
		cmd = "volume +5";

	// This makes changing volume kinda smooth and neat,
	// but it feels kinda screwy sometimes too.
//	u32 wpadheld = WPAD_ButtonsHeld(0);
//	if( (wpadheld & WPAD_BUTTON_MINUS) && g_status.volume > 0 )
//		cmd = "volume -1";
//	else if( (wpadheld & WPAD_BUTTON_PLUS) && g_status.volume < 100 )
//		cmd = "volume +1";
	
	if( !cmd.empty() ) {
		g_client.sendCommand(cmd);
		g_client.sendCommand("status");
	}

	return true;
}

static bool is_completion_error(const std::string& code)
{
	return code.empty() || code.size() >= 4 && code.substr(0, 4) == "ACK ";
}

static void parse_responses(const std::vector<std::string>& responses, std::map<std::string, std::string>& resp_map)
{
	for( std::vector<std::string>::const_iterator it = responses.begin(); it != responses.end(); ++it ) {
		const std::string& response = *it;
		std::string::size_type colon_pos = response.find(": ");
		if( colon_pos != std::string::npos ) {
			std::string key = response.substr(0, colon_pos);
			std::transform(key.begin(), key.end(), key.begin(), tolower);
			std::string value = response.substr(colon_pos + 2);
			resp_map.insert(std::make_pair(key, value));
		}
	}
}

static void process_reply_status(const std::vector<std::string>& responses)
{
	std::map<std::string, std::string> resp_map;
	parse_responses(responses, resp_map);

	g_status.isvalid = true;
	g_status.volume = atoi(resp_map["volume"].c_str());
	g_status.state = resp_map["state"];
	g_status.songid = atoi(resp_map["songid"].c_str());

	std::string time = resp_map["time"];
	std::string::size_type colon_pos = time.find(':');
	if( colon_pos != std::string::npos ) {
		g_status.time_elapsed = atoi(time.substr(0, colon_pos).c_str());
		g_status.time_total = atoi(time.substr(colon_pos+1).c_str());
	} else {
		g_status.time_elapsed = g_status.time_total = 0;
	}

	if( g_songinfo.isvalid && g_songinfo.songid != g_status.songid )
		g_songinfo.isvalid = false;
}

static void process_reply_playlistid(int songid, const std::vector<std::string>& responses)
{
	// We only care about playlist info if song matches one currently playing.
	if( !g_status.isvalid || songid != g_status.songid )
		return;

	std::map<std::string, std::string> resp_map;
	parse_responses(responses, resp_map);

	g_songinfo.isvalid = true;
	g_songinfo.songid = songid;
	g_songinfo.filename = resp_map["file"];
	g_songinfo.artist = resp_map["artist"];
	g_songinfo.title = resp_map["title"];
	g_songinfo.scroll = 0;
}

static void process_reply(const MPReply& reply)
{
	// For now, ignore all error replies.
	if( is_completion_error(reply.completion_code) )
		return;

	// Handle replies we care about.
	if( reply.command == "status" ) {
		process_reply_status(reply.responses);
	} else if( reply.command.size() > 11 && reply.command.substr(0, 11) == "playlistid " ) {
		int songid = atoi(reply.command.substr(11).c_str());
		process_reply_playlistid(songid, reply.responses);
	}
}

static void process_replies()
{
	MPReply reply;
	while( g_client.getReply(&reply) )
		process_reply(reply);
}

static std::string trim(const std::string& str)
{
	std::string::size_type p1 = 0;
	while( p1 < str.size() && (str[p1] == ' ' || str[p1] == '\t') )
		++p1;

	std::string::size_type p2 = str.size();
	while( p2 > 0 && (str[p2-1] == ' ' || str[p2-1] == '\t') )
		--p2;

	if( p1 == 0 && p2 == str.size() )
		return str;

	if( p2 <= p1 )
		return std::string();

	return str.substr(p1, p2-p1);
}

static bool setMPDHost()
{
	if( !fatInitDefault() )
		return false;
	
	FILE* f = fopen(CONF_FILENAME.c_str(), "r");
	if( !f )
		f = fopen(("/" + CONF_FILENAME).c_str(), "r");
	if( !f )
		return false;

	char line[1024];
	while( !feof(f) ) {
		fgets(line, sizeof(line), f);
		std::string linestr = line;

		// Strip newline.
		if( !linestr.empty() && linestr[linestr.size()-1] == '\n' )
			linestr.resize(linestr.size()-1);

		// Split into key and value.
		std::string::size_type equal_pos = linestr.find("=");
		if( equal_pos != std::string::npos ) {
			std::string key = trim(linestr.substr(0, equal_pos));
			std::string value = trim(linestr.substr(equal_pos+1));
			if( key == "MPD_HOST" ) {
				g_host = value;
				return true;
			}
		}
	}

	return false;
}

int main()
{
	GRRLIB_InitVideo();

	g_font = GRRLIB_LoadTexture(GRRLIB_font1);
	g_background = GRRLIB_LoadTexture(background);
	g_logo = GRRLIB_LoadTexture(logo);
	g_timebar = GRRLIB_LoadTexture(timebar);
	g_timepos_play = GRRLIB_LoadTexture(play);
	g_timepos_pause = GRRLIB_LoadTexture(pause);
	g_timepos_stop = GRRLIB_LoadTexture(stop);
	g_volumebar = GRRLIB_LoadTexture(volumebar);
	g_volumepos = GRRLIB_LoadTexture(volumepos);

	paint_first();

	s32 ret = WPAD_Init();
	if( ret < 0 )
		fail_exit("Initialize Wiimote failed!");
	WPAD_SetDataFormat(0, WPAD_FMT_BTNS);

	paint_netinit();
	ret = initialize_network();
	if( ret != 0 )
		fail_exit("Initialize network failed!");

	if( !setMPDHost() ) {
		fail_warn("MPD_HOST not set; using default", 5);
		g_host = DEFAULT_HOST;
	}

	if( !g_client.init(g_host, PORT) )
		fail_exit("Client initialize failed!");

	time_t tlaststatus = time(NULL);
	time_t tlastsonginfo = time(NULL);
	while( true ) {
		if( g_client.connected() ) {
			// Send "status" command once per second,
			// but only if no other responses are pending.
			time_t tnow = time(NULL);
			if( tnow - tlaststatus > 0 && !g_client.hasReply() ) {
				g_client.sendCommand("status");
				tlaststatus = tnow;
			}

			// Send "playlistid" command to get song info if current
			// song info is not valid with current status.
			tnow = time(NULL);
			if( g_status.isvalid && (!g_songinfo.isvalid || g_songinfo.songid != g_status.songid) &&
			    tnow - tlastsonginfo > 0 && !g_client.hasReply()  )
			{
				char cmd[80];
				snprintf(cmd, sizeof(cmd), "playlistid %d", g_status.songid);
				g_client.sendCommand(cmd);
				tlastsonginfo = tnow;
			}
		}

		// Process WiiMote buttons.
		if( WPAD_ScanPads() >= WPAD_ERR_NONE ) {
			if( !process_wpad_commands() )
				break;
		}

		if( g_client.connected() ) {
			process_replies();
		} else {
			g_status.isvalid = false;
			g_songinfo.isvalid = false;
		}

		paint();
		GRRLIB_Render();

		// Update song scroll position.
		if( g_songinfo.isvalid )
			g_songinfo.scroll++;
	}

	paint_exit();
	WPAD_Flush(WPAD_CHAN_ALL);
	WPAD_Disconnect(WPAD_CHAN_ALL);
	WPAD_Shutdown();
	GRRLIB_Stop();
	exit(0);
	return 0;
}

