// mpclient.cpp
// Copyright (c) 2008 Phillip Hellewell

#include "mpclient.h"
#include "utils.h"
#include <ogc/lwp.h>

static lwp_t g_comm_thread = 0;

MPClient::MPClient()
	: m_host()
	, m_port(0)
	, m_lineclient()
	, m_shutdown(false)
	, m_command_fifo()
	, m_reply_fifo()
{
	LWP_MutexInit(&m_command_fifo_mutex, false);
	LWP_MutexInit(&m_reply_fifo_mutex, false);
}

MPClient::~MPClient()
{
	m_shutdown = true;
	m_lineclient.disconnect();
}

bool MPClient::init(const std::string& host, int port)
{
	m_host = host;
	m_port = port;

	if( !m_lineclient.init() )
		return false;

	// Return if comm thread already initialized.
	if( g_comm_thread )
		return true;

	// Start the comm thread.
	return LWP_CreateThread(&g_comm_thread, static_commloop, this, NULL, 0, 80) >= 0;
}

void MPClient::sendCommand(const std::string& command)
{
	LWP_MutexLock(m_command_fifo_mutex);
	m_command_fifo.push(command);
	LWP_MutexUnlock(m_command_fifo_mutex);
}

bool MPClient::hasReply()
{
	return getReply(0);
}

bool MPClient::getReply(MPReply* reply)
{
	LWP_MutexLock(m_reply_fifo_mutex);
	bool ret = !m_reply_fifo.empty();
	if( ret && reply ) {
		*reply = m_reply_fifo.front();
		m_reply_fifo.pop();
	}
	LWP_MutexUnlock(m_reply_fifo_mutex);
	return ret;
}

// -----------------------------------------------------------------------

void* MPClient::static_commloop(void* mpclient)
{
	reinterpret_cast<MPClient*>(mpclient)->commloop();
	return NULL;
}

void MPClient::commloop()
{
	while( !m_shutdown ) {
		if( !m_lineclient.connected() ) {
			// Clear command buffer.
			LWP_MutexLock(m_command_fifo_mutex);
			m_command_fifo = std::queue<std::string>();
			LWP_MutexUnlock(m_command_fifo_mutex);
			// Clear reply buffer.
			LWP_MutexLock(m_reply_fifo_mutex);
			m_reply_fifo = std::queue<MPReply>();
			LWP_MutexUnlock(m_reply_fifo_mutex);

			// Connect and receive welcome message.
			if( m_lineclient.connect(m_host, m_port) )
				receive_reply();
			else
				millisleep(10);
		} else {
			// Get a command if there is one in the queue.
			std::string command;
			LWP_MutexLock(m_command_fifo_mutex);
			if( !m_command_fifo.empty() ) {
				command = m_command_fifo.front();
				m_command_fifo.pop();
			}
			LWP_MutexUnlock(m_command_fifo_mutex);

			// If there is a command, handle it; else delay for a little bit.
			if( !command.empty() && m_lineclient.sendLine(command) ) {
				MPReply reply;
				reply.command = command;
				reply.responses = receive_reply();
				if( !reply.responses.empty() ) {
					reply.completion_code = reply.responses.back();
					reply.responses.pop_back();
				}
				LWP_MutexLock(m_reply_fifo_mutex);
				m_reply_fifo.push(reply);
				LWP_MutexUnlock(m_reply_fifo_mutex);
			} else {
				millisleep(10);
			}
		}
	}
}

std::vector<std::string> MPClient::receive_reply()
{
	static const time_t TIMEOUT = 30;

	std::vector<std::string> responses;

	time_t tstart = time(NULL);
	while( true ) {
		// Don't let more than TIMEOUT seconds elapse trying to get a reply.
		time_t tnow = time(NULL);
		if( tnow - tstart > TIMEOUT )
			return std::vector<std::string>();

		// Small delay if no lines are ready.
		if( !m_lineclient.hasLine() ) {
			millisleep(1);
			if( !m_lineclient.connected() )
				return std::vector<std::string>();
			continue;
		}

		// Add line to response.
		std::string line;
		if( m_lineclient.getLine(&line) )
			responses.push_back(line);

		// We're done if we got "OK" or "ACK".
		if( line == "OK" || line.size() >= 3 && line.substr(0, 3) == "OK " ||
		                    line.size() >= 4 && line.substr(0, 4) == "ACK " )
		{
			break;
		}
	}

	return responses;
}

